(function(
  // Reliable reference to the global object (i.e. window in browsers).
  global,

  // Dummy constructor that we use as the .constructor property for
  // functions that return Generator objects.
  GeneratorFunction
) {
  var hasOwn = Object.prototype.hasOwnProperty;

  if (global.wrapGenerator) {
    return;
  }

  function wrapGenerator(innerFn, self) {
    return new Generator(innerFn, self || null);
  }

  global.wrapGenerator = wrapGenerator;
  if (typeof exports !== "undefined") {
    exports.wrapGenerator = wrapGenerator;
  }

  var GenStateSuspendedStart = "suspendedStart";
  var GenStateSuspendedYield = "suspendedYield";
  var GenStateExecuting = "executing";
  var GenStateCompleted = "completed";

  // Returning this object from the innerFn has the same effect as
  // breaking out of the dispatch switch statement.
  var ContinueSentinel = {};

  wrapGenerator.mark = function(genFun) {
    genFun.constructor = GeneratorFunction;
    return genFun;
  };

  // Ensure isGeneratorFunction works when Function#name not supported.
  if (GeneratorFunction.name !== "GeneratorFunction") {
    GeneratorFunction.name = "GeneratorFunction";
  }

  wrapGenerator.isGeneratorFunction = function(genFun) {
    var ctor = genFun && genFun.constructor;
    return ctor ? GeneratorFunction.name === ctor.name : false;
  };

  function Generator(innerFn, self) {
    var generator = this;
    var context = new Context();
    var state = GenStateSuspendedStart;

    function invoke(method, arg) {
      if (state === GenStateExecuting) {
        throw new Error("Generator is already running");
      }

      if (state === GenStateCompleted) {
        throw new Error("Generator has already finished");
      }

      while (true) {
        var delegate = context.delegate;
        if (delegate) {
          try {
            var info = delegate.generator[method](arg);

            // Delegate generator ran and handled its own exceptions so
            // regardless of what the method was, we continue as if it is
            // "next" with an undefined arg.
            method = "next";
            arg = void 0;

          } catch (uncaught) {
            context.delegate = null;

            // Like returning generator.throw(uncaught), but without the
            // overhead of an extra function call.
            method = "throw";
            arg = uncaught;

            continue;
          }

          if (info.done) {
            context[delegate.resultName] = info.value;
            context.next = delegate.nextLoc;
          } else {
            state = GenStateSuspendedYield;
            return info;
          }

          context.delegate = null;
        }

        if (method === "next") {
          if (state === GenStateSuspendedStart &&
              typeof arg !== "undefined") {
            // https://people.mozilla.org/~jorendorff/es6-draft.html#sec-generatorresume
            throw new TypeError(
              "attempt to send " + JSON.stringify(arg) + " to newborn generator"
            );
          }

          if (state === GenStateSuspendedYield) {
            context.sent = arg;
          } else {
            delete context.sent;
          }

        } else if (method === "throw") {
          if (state === GenStateSuspendedStart) {
            state = GenStateCompleted;
            throw arg;
          }

          if (context.dispatchException(arg)) {
            // If the dispatched exception was caught by a catch block,
            // then let that catch block handle the exception normally.
            method = "next";
            arg = void 0;
          }
        }

        state = GenStateExecuting;

        try {
          var value = innerFn.call(self, context);

          // If an exception is thrown from innerFn, we leave state ===
          // GenStateExecuting and loop back for another invocation.
          state = context.done
            ? GenStateCompleted
            : GenStateSuspendedYield;

          var info = {
            value: value,
            done: context.done
          };

          if (value === ContinueSentinel) {
            if (context.delegate && method === "next") {
              // Deliberately forget the last sent value so that we don't
              // accidentally pass it on to the delegate.
              arg = void 0;
            }
          } else {
            return info;
          }

        } catch (thrown) {
          if (method === "next") {
            context.dispatchException(thrown);
          } else {
            arg = thrown;
          }
        }
      }
    }

    generator.next = invoke.bind(generator, "next");
    generator.throw = invoke.bind(generator, "throw");
  }

  Generator.prototype.toString = function() {
    return "[object Generator]";
  };

  function Context() {
    this.reset();
  }

  Context.prototype = {
    constructor: Context,

    reset: function() {
      this.next = 0;
      this.sent = void 0;
      this.tryStack = [];
      this.done = false;
      this.delegate = null;

      // Pre-initialize at least 20 temporary variables to enable hidden
      // class optimizations for simple generators.
      for (var tempIndex = 0, tempName;
           hasOwn.call(this, tempName = "t" + tempIndex) || tempIndex < 20;
           ++tempIndex) {
        this[tempName] = null;
      }
    },

    stop: function() {
      this.done = true;

      if (hasOwn.call(this, "thrown")) {
        var thrown = this.thrown;
        delete this.thrown;
        throw thrown;
      }

      return this.rval;
    },

    keys: function(object) {
      return Object.keys(object).reverse();
    },

    pushTry: function(catchLoc, finallyLoc, finallyTempVar) {
      if (finallyLoc) {
        this.tryStack.push({
          finallyLoc: finallyLoc,
          finallyTempVar: finallyTempVar
        });
      }

      if (catchLoc) {
        this.tryStack.push({
          catchLoc: catchLoc
        });
      }
    },

    popCatch: function(catchLoc) {
      var lastIndex = this.tryStack.length - 1;
      var entry = this.tryStack[lastIndex];

      if (entry && entry.catchLoc === catchLoc) {
        this.tryStack.length = lastIndex;
      }
    },

    popFinally: function(finallyLoc) {
      var lastIndex = this.tryStack.length - 1;
      var entry = this.tryStack[lastIndex];

      if (!entry || !hasOwn.call(entry, "finallyLoc")) {
        entry = this.tryStack[--lastIndex];
      }

      if (entry && entry.finallyLoc === finallyLoc) {
        this.tryStack.length = lastIndex;
      }
    },

    dispatchException: function(exception) {
      var finallyEntries = [];
      var caught = false;

      if (this.done) {
        throw exception;
      }

      // Dispatch the exception to the "end" location by default.
      this.thrown = exception;
      this.next = "end";

      for (var i = this.tryStack.length - 1; i >= 0; --i) {
        var entry = this.tryStack[i];
        if (entry.catchLoc) {
          this.next = entry.catchLoc;
          caught = true;
          break;
        } else if (entry.finallyLoc) {
          finallyEntries.push(entry);
        }
      }

      while ((entry = finallyEntries.pop())) {
        this[entry.finallyTempVar] = this.next;
        this.next = entry.finallyLoc;
      }

      return caught;
    },

    delegateYield: function(generator, resultName, nextLoc) {
      this.delegate = {
        generator: generator,
        resultName: resultName,
        nextLoc: nextLoc
      };

      return ContinueSentinel;
    }
  };
}).apply(this, Function("return [this, function GeneratorFunction(){}]")());

var url = require('url');
var zlib = require('zlib');
var http = require('http');
var https = require('https');
var Netrc = require('netrc');
var saveTo = require('save-to');
var rawBody = require('raw-body');
var proxyAgent = require('proxy-agent');
var debug = require('debug')('cogent');
exports = module.exports = extend();
exports.extend = extend;

var redirectStatusCodes = [
  301,
  302,
  303,
  305,
  307,
];

var httpErrorStati = [
  502,
  503,
  504,
];

var httpErrorCodes = [
  'ECONNRESET',
  'ETIMEDOUT',
  'EADDRINFO',
  'ESOCKETTIMEDOUT',
];

function extend(defaults) {
  wrapGenerator.mark(cogent);
  defaults = defaults || {}
  defaults.retries = defaults.retries || 0;
  defaults.redirects = 'redirects' in defaults ? defaults.redirects : 3;
  defaults.timeout = defaults.timeout || 5000;
  defaults.method = defaults.method || 'GET';
  defaults.gunzip = defaults.gunzip === false ? false : true;
  defaults.netrc = Netrc(defaults.netrc);
  return cogent;

  function cogent(uri, options) {
    var redirects, timeout, netrc, headers, o, req, res, code, stream, securrrr, retries, auth, agent, gunzip, request;

    return wrapGenerator(function cogent$($ctx0) {
      while (1) switch ($ctx0.next) {
      case 0:
        request = function request(done) {
          var called = false

          // timeout
          // note: timeout is only for the response,
          // not the entire request
          var id = setTimeout(function () {
            debug('timeout exceeded for %s %s', o.method, o.href)
            var err = new Error('timeout of ' + timeout + 'ms exceeded.')
            err.url = o
            next(err)
          }, timeout)

          req = (securrrr ? https : http).request(o)
          req.once('response', next.bind(null, null))
          req.once('error', next)
          req.end()

          function next(err, res) {
            if (called) {
              // dump the stream in case there are any
              // to avoid any possible leaks
              if (res) res.resume()
              return
            }

            called = true
            clearTimeout(id)

            // kill the request, specifically for timeouts
            // to do: tests for this #7
            if (err) req.abort()

            if (retries && (
              (err && ~httpErrorCodes.indexOf(err.code)) ||
              (res && ~httpErrorStati.indexOf(res.statusCode))
            )) {
              debug('retrying %s %s', o.method, o.href)
              retries--
              request(done)
            } else {
              done(err, res)
            }
          }
        };

        // options type checking stuff
        if (options === true)
          options = { json: true }
        else if (typeof options === 'string')
          options = { destination: options }
        else if (!options)
          options = {}

        // setup defaults;

        redirects = options.redirects || defaults.redirects;
        timeout = options.timeout || defaults.timeout;
        netrc = options.netrc ? Netrc(options.netrc) : defaults.netrc;
        headers = options.headers = options.headers || {};
        headers['accept-encoding'] = headers['accept-encoding'] || 'gzip';
        headers['user-agent'] = headers['user-agent'] || 'https://github.com/cojs/cogent';

        if (options.json)
          headers.accept = 'application/json';
      case 9:
        if (!true) {
          $ctx0.next = 68;
          break;
        }

        // create the request options object
        o = url.parse(uri);

        securrrr = o.protocol === 'https:';
        o.headers = options.headers;

        o.method = options.method || defaults.method

        // auth, either by .auth or by .netrc;

        // auth, either by .auth or by .netrc
        if (options.auth || defaults.auth) {
          o.auth = options.auth || defaults.auth
        } else {
          auth = netrc[o.hostname];

          if (auth) o.auth =
              (auth.login || auth.user || auth.username)
            + ':'
            + (auth.pass || auth.password);
        }

        // setup agent or proxy agent
        if ('agent' in options) {
          o.agent = options.agent
        } else if (options.proxy || defaults.proxy) {
          agent = proxyAgent(options.proxy || defaults.proxy, securrrr);

          if (agent)
            o.agent = agent;
        } else if (defaults.agent != null) {
          o.agent = defaults.agent
        }

        // retries is on a per-URL-request basis
        retries = options.retries || defaults.retries;

        $ctx0.next = 19;
        return request;
      case 19:
        res = $ctx0.sent;

        code = res.statusCode

        // redirect;

        if (!(redirects-- && ~redirectStatusCodes.indexOf(code))) {
          $ctx0.next = 28;
          break;
        }

        debug('redirecting from %s to %s', uri, res.headers.location);
        uri = url.resolve(uri, res.headers.location);

        // dump this stream
        res.resume() // dump this stream;

        delete $ctx0.thrown;
        $ctx0.next = 9;
        break;
      case 28:
        if (!(o.method.toUpperCase() === 'HEAD')) {
          $ctx0.next = 35;
          break;
        }

        res.req = req;
        res.resume();
        $ctx0.rval = res;
        delete $ctx0.thrown;
        $ctx0.next = 68;
        break;
      case 35:
        gunzip = typeof options.gunzip === 'boolean'
          ? options.gunzip
          : defaults.gunzip;

        if (!(!gunzip && (options.buffer || options.string || options.json))) {
          $ctx0.next = 38;
          break;
        }

        throw new Error('must gunzip if buffering the response');
      case 38:
        if (res.headers['content-encoding'] === 'gzip' && gunzip) {
          stream = res.pipe(zlib.createGunzip())
          stream.res = res
          // pass useful response stuff
          stream.statusCode = code
          stream.headers = res.headers
          // forward errors
          res.on('error', stream.emit.bind(stream, 'error'))
          // forward close
          // creates the following error:
          // Assertion failed: (!write_in_progress_ && "write in progress"), function Close, file ../src/node_zlib.cc, line 99.
          // make: *** [test] Abort trap: 6
          // res.on('close', stream.close.bind(stream))
          // forward destroy
          stream.destroy = res.destroy.bind(res)
          res = stream
        } else {
          res.res = res
        }

        res.req = req

        // save the file;

        if (!(code === 200 && options.destination)) {
          $ctx0.next = 48;
          break;
        }

        $ctx0.next = 43;
        return saveTo(res, options.destination);
      case 43:
        res.destination = options.destination;
        $ctx0.rval = res;
        delete $ctx0.thrown;
        $ctx0.next = 68;
        break;
      case 48:
        if (!options.buffer) {
          $ctx0.next = 52;
          break;
        }

        $ctx0.next = 51;
        return rawBody(res);
      case 51:
        res.buffer = $ctx0.sent;
      case 52:
        if (!(options.string || options.json)) {
          $ctx0.next = 61;
          break;
        }

        if (!res.buffer) {
          $ctx0.next = 57;
          break;
        }

        $ctx0.t0 = res.buffer.toString('utf8');
        $ctx0.next = 60;
        break;
      case 57:
        $ctx0.next = 59;
        return rawBody(res, { encoding: options.string || true });
      case 59:
        $ctx0.t0 = $ctx0.sent;
      case 60:
        res.text = $ctx0.t0;
      case 61:
        // buffer the response as JSON
        if (options.json) try {
          res.body = JSON.parse(res.text)
        } catch (err) {
          debug('"%s" is not valid JSON', uri)
        }

        $ctx0.rval = res;
        delete $ctx0.thrown;
        $ctx0.next = 68;
        break;
      case 68:
      case "end":
        return $ctx0.stop();
      }
    }, this);
  }
}